﻿using System;
using System.Collections.Generic;
using System.Windows.Controls;
using System.Windows.Data;
using System.ComponentModel;
using System.Windows;

namespace SearchPath
{
    class Astar
    {
        private Map map;
        private int startCellIdx;
        private int endCellIdx;
        private ListView lvOpenList;
        private ListView lvClosedList;
        private List<Node> OpenItemSource = new List<Node>();
        private List<Node> ClosedItemSource = new List<Node>();

        public Astar(Map map, ListView lvOpenList, ListView lvClosedList, int startCellIdx, int endCellIdx)
        {
            this.map = map;
            this.lvOpenList = lvOpenList;
            this.lvClosedList = lvClosedList;
            this.startCellIdx = startCellIdx % (map.MaxColumns * map.MaxRows);
            this.endCellIdx = endCellIdx % (map.MaxColumns * map.MaxRows);

            Init();
        }

        public void Init()
        {
            OpenItemSource.Clear();
            ClosedItemSource.Clear();

            Node startNode = new Node(startCellIdx);
            startNode.TotalCost = startNode.HeuristicCost = CalcHeuristicCost(startCellIdx);

            OpenItemSource.Add(startNode);
            lvOpenList.ItemsSource = OpenItemSource;
            lvClosedList.ItemsSource = ClosedItemSource;

            CollectionView view = (CollectionView)CollectionViewSource.GetDefaultView(lvOpenList.ItemsSource);
            view.SortDescriptions.Add(new SortDescription("TotalCost", ListSortDirection.Ascending));
        }

        private int CalcHeuristicCost(int idx)
        {
            int startx = (idx % map.MaxColumns);
            int starty = (idx / map.MaxRows);
            int endx = (endCellIdx % map.MaxColumns);
            int endy = (endCellIdx / map.MaxRows);
            int dx = Math.Abs(startx - endx);
            int dy = Math.Abs(starty - endy);

            int cellCost = 10;
            //return DiagonalHeuristik(dx, dy, cellCost);
            return EuklidHeuristik(dx, dy, cellCost);
            //return ManhattenHeuristik(dx, dy, cellCost);
        }
        private int ManhattenHeuristik(int dx, int dy, int cost)
        {
            return (dx + dy) * cost;
        }

        private int DiagonalHeuristik(int dx, int dy, int cost)
        {
            return (int)(cost * (dx + dy) + (Math.Sqrt(2) * cost - 2 * cost) * Math.Min(dx, dy));
        }

        private int EuklidHeuristik(int dx, int dy, int cost)
        {
            return (int)(Math.Sqrt(dx * dx + dy * dy) * cost);
        }

        public void SingleStep()
        {
            if (lvOpenList.Items.IsEmpty)
            {
                MessageBox.Show("Target not found.");
                return;
            }

            Node actNode = (Node)lvOpenList.Items[0];
            OpenItemSource.Remove(actNode);
            ClosedItemSource.Add(actNode);

            Expand(actNode);

            CollectionViewSource.GetDefaultView(lvClosedList.ItemsSource).Refresh();
            CollectionViewSource.GetDefaultView(lvOpenList.ItemsSource).Refresh();
          
            if (actNode.MyIdx == endCellIdx)
            {
                MessageBox.Show("Target found.");
                map.DrawPath(startCellIdx, ClosedItemSource);
            }
        }

        private void Expand(Node actNode)
        {
            int actColumn = actNode.MyIdx % map.MaxColumns;
            int actRow = actNode.MyIdx / map.MaxColumns;
            // upper left
            int nextNodeIdx = actNode.MyIdx - map.MaxColumns - 1;
            if (actColumn > 0 && actRow > 0 &&
                IsNextNodeAllowed(actNode.FromIdx, nextNodeIdx))
                InsertNextNode(actNode, nextNodeIdx, false);
            // upper
            nextNodeIdx++;
            if (actRow > 0 &&
                IsNextNodeAllowed(actNode.FromIdx, nextNodeIdx))
                InsertNextNode(actNode, nextNodeIdx, true);
            // upper right
            nextNodeIdx++;
            if (actColumn < map.MaxColumns - 1 && actRow > 0 &&
                IsNextNodeAllowed(actNode.FromIdx, nextNodeIdx))
                InsertNextNode(actNode, nextNodeIdx, false);
            // left
            nextNodeIdx = actNode.MyIdx - 1;
            if (actColumn > 0 &&
                IsNextNodeAllowed(actNode.FromIdx, nextNodeIdx))
                InsertNextNode(actNode, nextNodeIdx, true);
            // right
            nextNodeIdx = actNode.MyIdx + 1;
            if (actColumn < map.MaxColumns - 1 &&
                IsNextNodeAllowed(actNode.FromIdx, nextNodeIdx))
                InsertNextNode(actNode, nextNodeIdx, true);
            // lower left
            nextNodeIdx = actNode.MyIdx + map.MaxColumns - 1;
            if (actColumn > 0 && actRow < map.MaxRows - 1 &&
                IsNextNodeAllowed(actNode.FromIdx, nextNodeIdx))
                InsertNextNode(actNode, nextNodeIdx, false);
            // lower
            nextNodeIdx++;
            if (actRow < map.MaxRows - 1 &&
                IsNextNodeAllowed(actNode.FromIdx, nextNodeIdx))
                InsertNextNode(actNode, nextNodeIdx, true);
            // lower right
            nextNodeIdx++;
            if (actColumn < map.MaxColumns - 1 && actRow < map.MaxRows - 1 &&
                IsNextNodeAllowed(actNode.FromIdx, nextNodeIdx))
                InsertNextNode(actNode, nextNodeIdx, false);
        }

        private bool IsNextNodeAllowed(int fromIdx, int nextNodeIdx)
        {
            if (fromIdx == nextNodeIdx) return false;
            if (!map.arrayCells[nextNodeIdx].walkable) return false;
            if (null != ClosedItemSource.Find(item => item.MyIdx == nextNodeIdx)) return false;
            return true;
        }

        private void InsertNextNode(Node actNode, int nextNodeIdx, bool orthogonal)
        {
            Node prevInsertedNode = OpenItemSource.Find(item => item.MyIdx == nextNodeIdx);
            if (null == prevInsertedNode)
            {
                Node nextNode = new Node(nextNodeIdx);
                nextNode.FromIdx = actNode.MyIdx;
                nextNode.PrevWayCost = actNode.PrevWayCost + map.GetFromToCost(actNode.MyIdx, nextNodeIdx, orthogonal);
                nextNode.HeuristicCost = CalcHeuristicCost(nextNodeIdx);
                nextNode.TotalCost = nextNode.PrevWayCost + nextNode.HeuristicCost;
                OpenItemSource.Add(nextNode);
            }
            else
            {
                int PrevWay = actNode.PrevWayCost + map.GetFromToCost(actNode.MyIdx, nextNodeIdx, orthogonal);
                if (PrevWay < prevInsertedNode.PrevWayCost)
                {
                    prevInsertedNode.PrevWayCost = PrevWay;
                    prevInsertedNode.FromIdx = actNode.MyIdx;
                    prevInsertedNode.TotalCost = prevInsertedNode.HeuristicCost + PrevWay; 
                }
            }
        }
    }
}
